#!/usr/bin/env python3
#==========================================================================
# (c) 2017 Total Phase, Inc.
#--------------------------------------------------------------------------
# Project : A2B Bus Monitor Sample Code
# File    : audio_capture.py
#--------------------------------------------------------------------------
# Auto-detection test routine
#--------------------------------------------------------------------------
# Redistribution and use of this file in source and binary forms, with
# or without modification, are permitted.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
# FOR A PARTICULAR PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL THE
# COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
# INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
# BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
# LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN
# ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.
#==========================================================================#
# This script requires Python 3.x
#==========================================================================#

#==========================================================================
# IMPORTS
#==========================================================================
import io
import sys
import time
import struct
import threading

from promira import Promira

#==========================================================================
# CLASS
#==========================================================================
# The AudioFile class abstracts the writing of the audio data to file.
class AudioFile:
    def __init__ (self, filename):
        self.filename = filename
        self.start_index = None
        self.offset_index = None

        # Output file
        # Output file is a 32-bit unsigned, mono, little-endian PCM file
        self.outfile = open(self.filename, 'wb')
        # Temporary buffer
        self.buffer  = io.BytesIO()
        self.fp = self.buffer

    # Keep track of the starting timestamp for this audio stream.
    def start (self, start):
        self.start_index = start

    def getStart (self):
        return self.start_index

    # Keep track of the offset.
    def offset (self, offset):
        self.offset_index = offset

    # Write audio stream to file or temporary buffer
    def write (self, data):
        # In the case of multiple streams, it is possible that each stream
        # will start at a different time.  In order to make analysis easier
        # each stream will be trimmed to the same start time. The offset
        # keeps track of the start time for all streams and the audio is
        # appropriately trimmed.
        #
        # Until an offset is set, the audio is stored in a temporary array
        # buffer.  Once an offset is set, the audio data is dumped to a
        # file.  After that, all audio is written directly to the file.
        if self.fp == self.buffer and self.offset_index != None:
            delta = self.offset_index - self.start_index

            # Truncate buffer and write all to outfile
            buffer_data = self.buffer.getvalue()
            self.outfile.write(buffer_data[delta:])

            # Switch fp
            self.fp = self.outfile

            # Close buffer
            self.buffer.close()
        # Write to fp
        self.fp.write(data)

    def close (self):
        self.outfile.close()

#==========================================================================
# CLASS
#==========================================================================
# The Stream class abstracts the WebSocket access to the Promira device
# to stream audio data.  The binary method processes the audio data as it
# arrives.
class Stream (Promira):
    def __init__ (self, url, probe, channel, rate, event):
        self.rate     = rate
        self.probe    = str(probe)
        self.channel  = channel
        self.filename = 'raw-audio-ch' + self.channel + '.pcm'
        self.event    = event

        self.start_ts = None
        self.last_ts = None

        self.audiofile = AudioFile(self.filename)

        super(Stream, self).__init__(url + "probe=" + self.probe + "&channel=" + self.channel)

    def binary (self, data):
        # The binary stream has a special protocol.  If the length of the data is
        # 8 bytes, then the data encodes the current timestamp of the audio stream.
        # If the data is more than 8 bytes, then the data encodes audio data.
        #
        # This is important because it is possible that there could be processing
        # problems which could result in dropped audio data.  In this case the audio
        # stream will send a new timestamp to inform the user that packets have been
        # dropped.  If a new timestamp is received, the audio stream needs to zeropad
        # the data for that length of time so that the audio data output file
        # correctly has the right amount of silence to keep all audio files
        # synchronized.

        # If length 8
        if len(data) == 8:
            ts = struct.unpack_from('d',data)[0]

            # If the first timestamp, store as the start
            if self.start_ts == None:
                self.start_ts = ts
                self.last_ts  = ts

                # Set the start time index in the audio file
                self.audiofile.start(int(ts * self.rate) * 4)

                # Once the channel has started, decrement the counter
                # in the event.
                self.event.nchannels -= 1

                # Once all channels have started, trigger the event.
                if self.event.nchannels == 0:
                    self.event.set()

            # This is not the first timestamp, which means that audio packets
            # have dropped.  Check difference between last timestamp and zeropad
            else:
                delta = ts - self.last_ts
                print('ch:', self.channel, 'zeropad:', delta)

                # Zeropad the delta time
                zeropad = bytearray(int(delta * self.rate) * 4)
                self.audiofile.write(zeropad)
                self.last_ts = ts

            print('ch:', self.channel, 'ts:', self.last_ts)

        # Handle normal audio data
        else:
            self.last_ts += len(data) / 4 / self.rate
            self.audiofile.write(data)
            print('ch:', self.channel, 'ts:', self.last_ts, 'stream:', len(data), 'bytes')

    def end (self):
        self.audiofile.close()

    def getBufferStart (self):
        return self.audiofile.getStart()

    def setOffset (self, offset):
        print("audiofile setOffset:", offset)
        self.audiofile.offset(offset)


#==========================================================================
# MAIN PROGRAM
#==========================================================================
if len(sys.argv) < 4:
    print('usage: example URL SAMPLE_RATE CH1 CH2 CHn...')
    print('URL must start with ws:// or ws://')
    print('SAMPLE_RATE is the audio sample rate in Hz, e.g. 48000')
    print('CH1... is a list of channels to record as integers')
    sys.exit(1)

url   = sys.argv[1]
rate  = int(sys.argv[2])
channels = sys.argv[3:]
streams  = []
probe = 0

# Connect to the Promira device
promira = Promira(url)
promira.connect()

# Create a session token for API access and subscribe to the WebSocket event stream.
session = promira.login([ 'admin', 'promira' ])
promira.subscribe()

# Start the capture.  Remember that this capture must be started before
# the Discovery process starts on the A2B bus of interest.
print('Starting capture...')
promira.a2b_capture_start()
promira.a2b_capture_audio(True)

# An event system is used to keep track of when all the audio threads have started.
print('Audio streaming...')
event = threading.Event()
event.nchannels = len(channels)

# Start all channels provided on the command line
for channel in channels:
    stream = Stream(url.replace('/api', '/audio?'), probe, channel, rate, event)
    stream.connect(session)
    streams.append(stream)

try:
    # Wait for all channels to start.
    event.wait()

    # Now that all channels have started, get the max start time for all streams
    starts = set()
    for stream in streams:
        starts.add(stream.getBufferStart())
    offset = max(starts)

    # Set the offset in all streams.
    for stream in streams:
        stream.setOffset(offset)

    # This is the operating loop.  To exit the loop when running the script, simply
    # stop the script (Ctrl+C)l
    while 1:
        time.sleep(1)

except:
    pass

print()
print('Stopping capture...')
promira.a2b_capture_stop()

print('Stopping streaming...')
for stream in streams:
    stream.end()
    stream.disconnect()

print('Disconnecting...')
promira.disconnect()

print('Done.')

sys.exit(0)
